// NavMerge will merge any number of Rinex nav files into a unique superset,
// and either write them out to a new Rinex file (if an output file is given),
// or write a summary of the data to the screen. NavMerge also finds
// and fixes full week number when it is inconsistent with epoch.
//------------------------------------------------------------------------------------
#pragma ident "$Id: //depot/sgl/gpstk/dev/apps/Rinextools/NavMerge.cpp#1 $"

/**
 * @file NavMerge.cpp
 * Read, summarize and optionally merge, Rinex navigation files.
 */

//------------------------------------------------------------------------------------
//============================================================================
//
//  This file is part of GPSTk, the GPS Toolkit.
//
//  The GPSTk is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or
//  any later version.
//
//  The GPSTk is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with GPSTk; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  
//  Copyright 2004, The University of Texas at Austin
//
//============================================================================

//------------------------------------------------------------------------------------
#include "RinexNavData.hpp"
#include "RinexNavHeader.hpp"
#include "RinexNavStream.hpp"
#include "BCEphemerisStore.hpp"
//------------------------------------------------------------------------------------

using namespace std;

// Returns 0 when successful.
int main(int argc, char *argv[])
{
   if (argc<2)
   {
      cout << "Usage: NavMerge [-o<Output Rinex nav file>] <Rinex nav file...>\n";
      cout << "         If output file is omitted, a summary is produced.\n";
      cout << "         Also correct the ephemeris when the GPS full week number\n";
      cout << "         is not consistent with the ephemeris epoch.\n";
      return -1;
   }

   try
   {
      int i=1;
      string filename,outfile;
      while(i < argc) {
         filename = argv[i];
         if(filename[0]=='-' && (filename[1]=='o' || filename[1]=='O')) {
            outfile = filename.substr(2);
            cout << "Output file name is " << outfile << endl;
            argv[i][0] = '\0';
            i++;
         }
         i++;
      }

      gpstk::RinexNavHeader rnh,rnhout;
      gpstk::RinexNavData rne;
      gpstk::BCEphemerisStore EphStore;
      gpstk::RinexNavStream RNFileOut;

      if(outfile != string("")) {
         RNFileOut.open(outfile.c_str(),ios::out);
         RNFileOut.exceptions(fstream::failbit);
         rnhout.version = 2.1;
         rnhout.valid |= gpstk::RinexNavHeader::versionValid;
         rnhout.fileType = string("NAVIGATION");
         rnhout.fileProgram = string("NavMerge");
         rnhout.fileAgency = string("GPSTK");
         rnhout.valid |= gpstk::RinexNavHeader::runByValid;
         rnhout.commentList.clear();
         rnhout.valid |= gpstk::RinexNavHeader::commentValid;
         rnhout.valid |= gpstk::RinexNavHeader::endValid;
      }

      int na=1,n=0,nf;
      while(na < argc) {
         filename = argv[na];
         if(filename == string("")) { na++; continue; }

         try {
            gpstk::RinexNavStream RNFileIn(filename.c_str());
            RNFileIn.exceptions(fstream::failbit);
      
            RNFileIn >> rnh;
            if(rnh.valid & gpstk::RinexNavHeader::ionAlphaValid) {
               for(i=0; i<4; i++) rnhout.ionAlpha[i]=rnh.ionAlpha[i];
               rnhout.valid |= gpstk::RinexNavHeader::ionAlphaValid;
            }
            if(rnh.valid & gpstk::RinexNavHeader::ionBetaValid) {
               for(i=0; i<4; i++) rnhout.ionBeta[i]=rnh.ionBeta[i];
               rnhout.valid |= gpstk::RinexNavHeader::ionBetaValid;
            }
            if(rnh.valid & gpstk::RinexNavHeader::deltaUTCValid) {
               rnhout.A0 = rnh.A0;
               rnhout.A1 = rnh.A1;
               rnhout.UTCRefWeek = rnh.UTCRefWeek;
               rnhout.UTCRefTime = rnh.UTCRefTime;
               rnhout.valid |= gpstk::RinexNavHeader::deltaUTCValid;
            }
            if(rnh.valid & gpstk::RinexNavHeader::leapSecondsValid) {
               rnhout.leapSeconds = rnh.leapSeconds;
               rnhout.valid |= gpstk::RinexNavHeader::leapSecondsValid;
            }

            nf = 0;
            while (RNFileIn >> rne)
            {
               nf++;
               n++;
                  // check that week number (associated with HOW)
                  // is consistent with TOC.
                  // (NB. in Rinex nav file, the week number is
                  // associated with the TOE; gpstk::RinexNavData
                  // converts it to associate with the HOW)
               int wkTOC,wk;
               wk = rne.weeknum;                // 'weeknum' associated with HOW
               wkTOC = rne.time.GPSfullweek();  // 'time' comes from epoch line
               if(ABS(wk-wkTOC) > 1) {          // HOW and TOC should be w/in 1 week
                  double dt = double(wk-wkTOC)/1024.0;
                  dt += (dt < 0.0 ? -0.5 : 0.5);
                  wk -= int(dt) * 1024;
                  if(ABS(wk-wkTOC) > 1) {
                     cout << "WARNING: Ephemeris in " << filename
                        << " for satellite G"
                        << setw(2) << setfill('0') << rne.PRNID << setfill(' ')
                        << " at time " << rne.time
                        << " has inconsistent week number " << rne.weeknum << endl;
                  }
                  else {
                     cout << "NavMerge corrected the week in G"
                        << setw(2) << setfill('0') << rne.PRNID << setfill(' ')
                        << " " << rne.time
                        << " " << filename << endl;
                     rne.weeknum = wk;
                  }
               }
                  // add to the store
               EphStore.addEphemeris(rne);
            }
         }
         catch(gpstk::Exception& e) {
            cout << "Exception: " << e << endl;
         }
         na++;
         cout << "Read " << setw(4) << nf << " ephemerides from file "
            << filename << endl;
      }
      cout << "Read " << setw(4) << n << " total ephemerides." << endl;

         // pull out all the ephemerides
      list<gpstk::EngEphemeris> EphList;
      i = EphStore.addToList(EphList);

      if(outfile != string("")) {
            // write the output header
         RNFileOut << rnhout;

            // write out all the ephemerides
         list<gpstk::EngEphemeris>::iterator it=EphList.begin();
         n=0;
         while(it != EphList.end()) {
            rne = gpstk::RinexNavData(*it);
            n++;
            RNFileOut << rne;
            it++;
         }
         cout << "Wrote " << setw(3) << n << " unique ephemerides to file "
            << outfile << endl;
      }
      else {
         EphStore.dump(1);
      }

      return 0;
   }
   catch(gpstk::Exception& e)
   {
      cout << e;
      return 1;
   }
   catch (...)
   {
      cout << "unknown error.  Done." << endl;
      return 1;
   }

   return 0;
}
