#pragma ident "$Id: //depot/sgl/gpstk/dev/apps/reszilla/reszilla.cpp#4 $"
//
// Compute observed range deviations (ORDs) and double
// difference residuals from rinex obs files.
//

#include <fstream>

#include "readers.hpp"
#include "util.hpp"

#include <BCEphemerisStore.hpp>

using namespace std;

int verbosity=1;

int main(int argc, char *argv[])
{
   string hmsFmt="%Y %3j %02H:%02M:%02S";
   string sodFmt="%Y %3j %7.1s";
   string ordMode="c1p2";
   string resMode="none";
   string clkMode="epoch";
   string ddMode="sv";
   unsigned long msid=0;
   double sigmaMask=6;
   timeFormat = hmsFmt;

   try
   {
      gpstk::CommandOptionNoArg
         nearOption('n', "search-near", "Use BCEphemeris.searchNear()"),
         helpOption('h', "help", "Print help usage"),
         clkOption('c', "ignore-clock", "Ignore any Rx clock that is present in the rinex obs data."),
         svTimeOption('\0', "svtime", "Observation data is in SV time frame. Default is RX time frame."),
         statsOption('s', "no-stats", "Don't compute & output the statistics");
   
      gpstk::CommandOptionWithAnyArg
         rawOutputOption('r', "raw-output", "Dump the computed residuals/ords into specified file. If '-' is given as the file name, output is sent to stdout. The default is to no otput the raw residuals."),
         timeFmtOption('t', "timeFormat", "Daytime format specifier used for the timestamps in the raw output. The default is \"" + timeFormat + "\". If this option is specified with the format as \"s\", the format \"" + sodFmt + "\" is used."),
         ordModeOption('\0', "omode", "ORD mode: none, P1P2, C1P2, C1, P1, P2. Default is " + ordMode),
         resModeOption('\0', "rmode", "Residual mode: none, P1L1, C1L1, P2L2. Default is " + resMode + "."),
         clkModeOption('\0', "cmode", "Clock computation mode: epoch, linear. Default is " + clkMode + "."),
         ddModeOption('\0', "ddmode", "Double difference residual mode: none, SV, C1P2. Default is " + ddMode + "."),
         mscFileOption('\0', "msc", "Station coordinate file"),
         ephFileOption('e', "ephemeris", "Ephemeris data file name (either broadcast in RINEX nav, broadcast in FIC, or precise in SP3)", false),
         obsFileOption('o', "observation", "Observation data file name", true),
         metFileOption('w', "weather", "Weather data file name");

      gpstk::CommandOptionWithNumberArg
         msidOption('m', "msid", "Station to process data for. Used to select a station from smooted data files."),
         sigmaOption('\0', "sigma", "Multiplier for sigma stripping used in computation of statistics on the raw residuals. The default is " + gpstk::StringUtils::asString((int)sigmaMask) + "."),
         verbosityOption('v', "verbosity", "How much detail do you want about intermediate steps?");

      string appDesc("Computes various residuals from GPS observations.");
      gpstk::CommandOptionParser cop(appDesc);
      cop.parseOptions(argc, argv);

      if (helpOption.getCount() || cop.hasErrors())
      {
         if (cop.hasErrors() && helpOption.getCount()==0)
         {
            cop.dumpErrors(cout);
            cout << "use -h for help" << endl;
         }
         else
         {
            cop.displayUsage(cout);
            if (helpOption.getCount() > 1)
               cout << endl
                    << "Verbosity values:" << endl
                    << "  0: nothing but the results" << endl
                    << "  1: Output status before potentially time consuming operations" << endl
                    << "  2: more details about each step and the options chosen" << endl
                    << "  3: add the reasons for editing data" << endl
                    << "  4: dump intermediate values" << endl
                    << endl
                    << "Example command to compute ORDs on an ICD-GPS-211 formated smoothed" << endl
                    << "measurement data file:" << endl
                    << "  reszilla --omode=p1 --svtime --msc=mscoords.cfg -m 85401 -o asm2004.138 -e s011138a.04n" << endl
                    << endl;

         }
         exit(0);
      }

      if (verbosityOption.getCount())
         verbosity = gpstk::StringUtils::asInt(verbosityOption.getValue()[0]);

      if (ddModeOption.getCount())
         ddMode = gpstk::StringUtils::lowerCase(ddModeOption.getValue().front());
      
      if (ordModeOption.getCount())
         ordMode=gpstk::StringUtils::lowerCase(ordModeOption.getValue()[0]);
      
      bool svTime = svTimeOption.getCount()>0;

      if (clkModeOption.getCount())
         clkMode=gpstk::StringUtils::lowerCase(clkModeOption.getValue().front());

      if (msidOption.getCount())
         msid=gpstk::StringUtils::asUnsigned(msidOption.getValue().front());

      if (sigmaOption.getCount())
         sigmaMask=gpstk::StringUtils::asDouble(sigmaOption.getValue().front());

      if (resModeOption.getCount())
         resMode=gpstk::StringUtils::lowerCase(resModeOption.getValue().front());

      if (timeFmtOption.getCount())
      {
         if ((timeFmtOption.getValue())[0] == "s")
            timeFormat = sodFmt;
         else
            timeFormat = (timeFmtOption.getValue())[0];
      }

      // set up where the raw data will be written
      ofstream ofs;
      string outputFileName;
      if (rawOutputOption.getCount())
      {
         outputFileName=(rawOutputOption.getValue())[0];
         if (outputFileName != "-")
            ofs.open(outputFileName.c_str());
      }

      if (verbosity>1)
      {
         cout << "Double difference mode: " << ddMode << endl
              << "Format to use for time in raw output: " << timeFormat << endl
              << "Data time tag: " << (svTime?"sv":"rx") << endl
              << "Clock computation mode: " << clkMode << endl
              << "Residual computation mode: " << resMode << endl
              << "Double difference computation mode: " << ddMode << endl
              << "Sigma stripping multiplier: " << sigmaMask << endl
              << "Verbosity: " << verbosity << endl;
         if (msid)
         {
            cout << "msid: " << msid << endl;
            if (mscFileOption.getCount()>0)
               cout << "msc file: " << mscFileOption.getValue().front() << endl;
         }
         if (rawOutputOption.getCount())
            cout << "Raw output file: " << outputFileName << endl;

      }

      // -------------------------------------------------------------------
      // End of processing/checking command line arguments, now on to the
      // data processing portion. First we get all the data into memory.
      // -------------------------------------------------------------------

      // Get the ephemeris data
      gpstk::EphemerisStore& eph = read_eph_data(ephFileOption);
      if (nearOption.getCount())
      {
         gpstk::BCEphemerisStore& bce = dynamic_cast<gpstk::BCEphemerisStore&>(eph);
         bce.SearchNear();
      }

      gpstk::WxObsData& wod = read_met_data(metFileOption);

      // Get the obs data from the first obs file
      string ofn = (obsFileOption.getValue())[0];
      RODEpochMap rem1;
      gpstk::RinexObsHeader roh1;
      read_obs_data(ofn, msid, rem1, roh1);

      // If a msid & msc file is specified, then get & use the msc file
      // to overwrite the position in the rinex obs header
      if (msid && mscFileOption.getCount()>0)
      {
         string mscfn = (mscFileOption.getValue())[0];
         read_msc_data(mscfn, msid, roh1);
      }

      if (ordMode!="none")
      {
         // Compute the ords
         ORDEpochMap oem1;
         compute_ords(oem1, rem1, roh1, eph, wod, svTimeOption, ordMode, clkMode);

         // Now, output statistics to stdout
         if (statsOption.getCount()==0)
            dumpStats(oem1, ordMode, sigmaMask);
         
         // Save the raw ORDs to a file
         if (rawOutputOption.getCount())
            if (outputFileName != "-")
               dump(ofs, oem1);
            else
               dump(cout, oem1);
      }

      if (resMode!="none")
      {
         PrnElevationMap pem = elevation_map(rem1, roh1, eph);

         // Compute the residuals

         // obviously to be done later

         // Now, output statistics to stdout
         
         // Save the raw ORDs to a file
      }

      // Now compute double difference residuals on remaining obs data
      for (int i=1; i<obsFileOption.getCount(); i++)
      {
         PrnElevationMap pem = elevation_map(rem1, roh1, eph);

         string fn=(obsFileOption.getValue())[i];
         RODEpochMap rem2;
         rem2.clear();
         gpstk::RinexObsHeader roh2;
         read_obs_data(fn, msid, rem2, roh2);

         if (verbosity)
            cout << "Computing double difference residuals from " << ofn
                 << " and " << fn << endl;

         if (ddMode == "sv")
         {

            // we need the rx clock offset for this double differece
            // Compute the ords just to get a clock offset between receivers
            if (clkOption.getCount())
            {
               ORDEpochMap oem1,oem2;
               compute_ords(oem1, rem1, roh1, eph, wod, svTimeOption, ordMode, clkMode);
               compute_ords(oem2, rem2, roh2, eph, wod, svTimeOption, ordMode, clkMode);
               add_clock_to_rinex(rem1, oem1);
               add_clock_to_rinex(rem2, oem2);
            }

            DDEpochMap ddem;
            computeDD(rem1, rem2, 10, ddem, pem);
               
            CycleSlipList sl1,sl2;
            sl1 = debiasPhase(L1, ddem);
            sl2 = debiasPhase(L2, ddem);
               
            if (statsOption.getCount()==0)
            {
               dumpStats(ddem, pem);
               dumpStats(sl1, L1, pem);
               dumpStats(sl1, L2, pem);
            }
            if (rawOutputOption.getCount())
               if (outputFileName != "-")
                  dump(ofs, ddem, L1, pem);
               else
                  dump(cout, ddem, L1, pem);
               
         }
         else if (ddMode=="c1p2")
         {
            DD2EpochMap ddem;
            computeDD(rem1, rem2, ddem);
            if (statsOption.getCount()==0)
            {
               dumpStats(ddem, pem);
            }
            if (rawOutputOption.getCount())
               if (outputFileName != "-")
                  dump(ofs, ddem, pem);
               else
                  dump(cout, ddem, pem);
            
         }
      } // end for()
   }
   catch (gpstk::Exception& e)
   {
      cerr << "Caught Excption: " << typeid(e).name() << endl
           << "Terminating." << endl;
      exit(0);
   }
   exit(0);
}
