#pragma ident "$Id: //depot/sgl/gpstk/dev/apps/reszilla/util.cpp#7 $"

#include "util.hpp"

#include "SP3EphemerisStore.hpp"
#include "RinexEphemerisStore.hpp"

using namespace std;


const RinexObsType& L1=gpstk::RinexObsHeader::L1;
const RinexObsType& L2=gpstk::RinexObsHeader::L2;
const RinexObsType& P1=gpstk::RinexObsHeader::P1;
const RinexObsType& P2=gpstk::RinexObsHeader::P2;
const RinexObsType& C1=gpstk::RinexObsHeader::C1;
const RinexObsType& D1=gpstk::RinexObsHeader::D1;
const RinexObsType& D2=gpstk::RinexObsHeader::D2;

string timeFormat;

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void compute_ords(ORDEpochMap& oem,
                  const RODEpochMap& rem,
                  const gpstk::RinexObsHeader& roh,
                  const gpstk::EphemerisStore& eph,
                  const gpstk::WxObsData& wod,
                  bool svTime,
                  const string& ordModeStr,
                  const std::string& clkModelStr)
{
   bool dualFreq=false;
   RinexObsType p1,p2;

   if      (ordModeStr=="p1p2" || ordModeStr=="c1p2") dualFreq=true;

   if      (ordModeStr=="p1p2") p1=P1, p2=P2;
   else if (ordModeStr=="c1p2") p1=C1, p2=P2;
   else if (ordModeStr=="c1")   p1=C1;
   else if (ordModeStr=="p1")   p1=P1;
   else if (ordModeStr=="p2")   p1=P2;
   else
   {
      cout << "Unknown ORD computation requested, mode=" << ordModeStr << endl;
      return;
   }

   if (verbosity)
      cout << "Computing observed range deviations." << endl;

   if (verbosity>1)
      if (svTime)
         cout << "Assuming data is tagged in SV time (time of emission)." << endl;
      else
         cout << "Assuming data is taged in Receiver time (gps time)." << endl;


   if (roh.antennaPosition[0] == 0 && roh.antennaPosition[1] == 0 && roh.antennaPosition[2] == 0)
   {
      cout << "Obs file doesn't appear to have a vaild antenna pos." << endl;
      return;
   }
   
   gpstk::ObsClockModel* cm;
   if (clkModelStr=="epoch")
   {
      if (verbosity)
         cout << "Using an epoch clock model." << endl;
      cm = new(gpstk::EpochClockModel);
   }
   else if (clkModelStr=="linear")
   {
      if (verbosity)
         cout << "Using a linear clock model." << endl;
      cm = new(gpstk::LinearClockModel);
   }
   else
   {
      cout << "Unknown clock model requestd, model=" << clkModelStr << endl;
      return;
   }

   cm->setElevationMask(5);
   gpstk::GPSGeoid gm;
   gpstk::ECEF ap(roh.antennaPosition);
   gpstk::Geodetic geo(ap, &gm);
   gpstk::NBTropModel tm(geo.getAltitude(), geo.getLatitude(), roh.firstObs.DOYday());

   RODEpochMap::const_iterator i;
   for (i=rem.begin(); i!=rem.end(); i++)
   {
      const gpstk::DayTime& t = i->first;
      const gpstk::RinexObsData& rod = i->second;

      gpstk::ORDEpoch& oe = oem[t];
      oe.time = t;

      // Now set up our trop model for this epoch
      const gpstk::WxObservation wx = wod.getMostRecent(t);
      if (verbosity>3)
         cout << wx << endl;
      if (wx.isAllValid())
         tm.setWeather(wx.temperature, wx.pressure, wx.humidity);

      // Walk over all prns in this epoch
      gpstk::RinexObsData::RinexPrnMap::const_iterator rpi;
      for (rpi=rod.obs.begin(); rpi!=rod.obs.end(); rpi++)
      {
         short prn = rpi->first.prn;

         gpstk::RinexObsData::RinexObsTypeMap rotm = rpi->second;

         try
         {
            if (dualFreq)
            {
               gpstk::ObsRngDev ord(rotm[p1].data, rotm[p2].data,
                                    prn, t, ap, eph, gm, tm, svTime);
               oe.ords[prn] = ord;
            }
            else
            {
               gpstk::ObsRngDev ord(rotm[p1].data, 
                                    prn, t, ap, eph, gm, tm, svTime);
               oe.ords[prn] = ord;
            }
         }
         catch (gpstk::EphemerisStore::NoEphemerisFound& e)
         {
            if (verbosity>2)
               cout << e.getText() << endl;
         }
      }

      cm->addEpoch(oe);
      if (verbosity>3)
         cout << *cm;
      oe.applyClockModel(*cm);
      if (verbosity>3)
         cout << oe;
   }
} // end of compute_ords()



// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void computeDD(const RODEpochMap& rx1,
               const RODEpochMap& rx2,
               double minMasterElevation,
               DDEpochMap& ddem,
               PrnElevationMap& pem)
{
   if (verbosity)
      cout << "Computing 2nd differences residuals across master SV." << endl;
   if (verbosity>1)
      cout << "Using a minimum master SV elevation of "
           << minMasterElevation << " degrees." << std::endl;

   gpstk::RinexPrn masterPrn(-1,gpstk::systemGPS);

   // We use the data from rx1 walk us through the data
   // loop over all epochs for this station
   RODEpochMap::const_iterator ei1, ei2;
   for (ei1=rx1.begin(); ei1!=rx1.end(); ei1++)
   {
      // first make sure we have data from the other receiver for this
      // epoch...
      const gpstk::DayTime& t = ei1->first;
      ei2 = rx2.find(t);
      if (ei2 == rx2.end())
      {
         if (verbosity>2)
            cout << "Epoch with no data in second file at " << t << endl;
         continue;
      }

      gpstk::RinexObsData e1 = ei1->second;
      gpstk::RinexObsData e2 = ei2->second;

      // Now to select a master PRN for this epoch.
      // First see if we can keep using the previous master
      gpstk::RinexPrn prn = masterPrn;
      gpstk::RinexObsData::RinexPrnMap::iterator oi1, oi2;
      oi1 = e1.obs.find(prn);
      oi2 = e2.obs.find(prn);
      if (oi1!=e1.obs.end() && oi2!=e2.obs.end() 
          && pem[t][prn.prn] > minMasterElevation)
      {
         // Yea, we can keep using the previous master
      }
      else
      {
         // Nope, we need to find a new master.
         // Criteria:  it has an elevation > the min, 
         //            it it on the way up (i.e. doppler>0),
         //            and there is a record for it on the other receiver
         for (oi1=e1.obs.begin(); oi1!=e1.obs.end(); oi1++)
         {
            prn = oi1->first;
            gpstk::RinexObsData::RinexObsTypeMap& obs1 = oi1->second;
            oi2 = e2.obs.find(prn);

            if (pem[t][prn.prn] > minMasterElevation
                && obs1[D1].data >= 0
                && oi2 != e2.obs.end())
            {
               masterPrn = prn;
               break;
            }
         }

         // If we can't find an acceptable master, we won't do a double difference on this epoch
         if (oi1==e1.obs.end())
         {
            if (verbosity>2)
               cout << "Could not find an SV with an elevation greater than "
                    << minMasterElevation << " degrees at "
                    << t << ".  Skipping epoch." << endl;
            continue;
         }

         if (verbosity>1)
            cout << "New master Prn: " << masterPrn << " at " << t << endl;
      }
      
      DDEpoch dde;
      if (dde.compute(e1, e2, masterPrn))
         ddem[t] = dde;
   }
} // end of computeDD()


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
bool DDEpoch::compute(gpstk::RinexObsData rx1,
                      gpstk::RinexObsData rx2,
                      gpstk::RinexPrn aMasterPrn)
{
   masterPrn = aMasterPrn;
   clockOffset = rx1.clockOffset - rx2.clockOffset;
   if (abs(clockOffset)>2.1e-3)
   {
      if (verbosity>2)
         cout << "Rx1-Rx2 clock offset of " << 1e3*clockOffset << " ms at "
              << " at " << rx1.time << endl; 
      return false;
   }

   map<RinexObsType, double> masterDiff, otherDiff;
   gpstk::RinexObsData::RinexPrnMap::iterator oi1, oi2;
   oi1 = rx1.obs.find(masterPrn);
   oi2 = rx2.obs.find(masterPrn);
   gpstk::RinexObsData::RinexObsTypeMap& rx1obs = oi1->second;
   gpstk::RinexObsData::RinexObsTypeMap& rx2obs = oi2->second;

   const gpstk::GPSGeoid gm;
   double a = gm.c()/gpstk::L1_FREQ;
   double b = gm.c()/gpstk::L2_FREQ;
   double coc = clockOffset*rx1obs[D1].data*a;  // clock offset correction

   masterDiff[C1] = rx1obs[C1].data   - (rx2obs[C1].data   + coc);
   masterDiff[P1] = rx1obs[P1].data   - (rx2obs[P1].data   + coc);
   masterDiff[L1] = rx1obs[L1].data*a - (rx2obs[L1].data*a + coc);
   masterDiff[D1] = rx1obs[D1].data*a - (rx2obs[D1].data*a + coc);
   masterDiff[P2] = rx1obs[P2].data   - (rx2obs[P2].data   + coc);
   masterDiff[L2] = rx1obs[L2].data*b - (rx2obs[L2].data*b + coc);
   masterDiff[D2] = rx1obs[D2].data*b - (rx2obs[D2].data*b + coc);

   // Now walk through all prns in track
   for (oi1=rx1.obs.begin(); oi1!=rx1.obs.end(); oi1++)
   {
      gpstk::RinexPrn prn = oi1->first;
      if (prn == masterPrn)
         continue;
         
      oi2 = rx2.obs.find(prn);
      if (oi2 == rx2.obs.end())
         continue;

      // Now compute the single difference for the 'other' SV
      rx1obs = oi1->second;
      rx2obs = oi2->second;
      coc = clockOffset*rx2obs[D1].data*a;

      otherDiff[C1] = rx1obs[C1].data   - (rx2obs[C1].data   + coc);
      otherDiff[P1] = rx1obs[P1].data   - (rx2obs[P1].data   + coc);
      otherDiff[L1] = rx1obs[L1].data*a - (rx2obs[L1].data*a + coc);
      otherDiff[D1] = rx1obs[D1].data*a - (rx2obs[D1].data*a + coc);
      otherDiff[P2] = rx1obs[P2].data   - (rx2obs[P2].data   + coc);
      otherDiff[L2] = rx1obs[L2].data*b - (rx2obs[L2].data*b + coc);
      otherDiff[D2] = rx1obs[D2].data*b - (rx2obs[D2].data*b + coc);

      // Now compute the double differences
      res[prn].res[C1] = masterDiff[C1]  - otherDiff[C1];
      res[prn].res[P1] = masterDiff[P1]  - otherDiff[P1];
      res[prn].res[L1] = masterDiff[L1]  - otherDiff[L1];
      res[prn].res[D1] = masterDiff[D1]  - otherDiff[D1];
      res[prn].res[P2] = masterDiff[P2]  - otherDiff[P2];
      res[prn].res[L2] = masterDiff[L2]  - otherDiff[L2];
      res[prn].res[D2] = masterDiff[D2]  - otherDiff[D2];

   }

   return true;
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void computeDD(const RODEpochMap& rx1,
               const RODEpochMap& rx2,
               DD2EpochMap& ddem)
{
   if (verbosity)
      cout << "Computing 2nd differences residuals across codes." << endl;

   // We use the data from rx1 walk us through the data
   // loop over all epochs for this station
   RODEpochMap::const_iterator ei1, ei2;
   for (ei1=rx1.begin(); ei1!=rx1.end(); ei1++)
   {
      // first make sure we have data from the other receiver for this
      // epoch...
      const gpstk::DayTime& t = ei1->first;
      ei2 = rx2.find(t);
      if (ei2 == rx2.end())
      {
         if (verbosity>2)
            cout << "Epoch with no data in second file at " << t << endl;
         continue;
      }

      gpstk::RinexObsData e1 = ei1->second;
      gpstk::RinexObsData e2 = ei2->second;

      DD2Epoch dde;
      if (dde.compute(e1, e2))
         ddem[t] = dde;
   }
} // end of computeDD()


// ---------------------------------------------------------------------
// Compute a double difference for each prn in track on both receivers
// (C1(prn1)-P2(prn1)) - (C1(prn2)-P2(prn2))
// ---------------------------------------------------------------------
bool DD2Epoch::compute(gpstk::RinexObsData rx1,
                       gpstk::RinexObsData rx2)
{
   gpstk::RinexObsData::RinexPrnMap::iterator oi1, oi2;
   for (oi1=rx1.obs.begin(); oi1!=rx1.obs.end(); oi1++)
   {
      gpstk::RinexPrn prn = oi1->first;
      oi2 = rx2.obs.find(prn);
      if (oi2 == rx2.obs.end())
         continue;

      gpstk::RinexObsData::RinexObsTypeMap& rx1obs = oi1->second;
      gpstk::RinexObsData::RinexObsTypeMap& rx2obs = oi2->second;

      // Now compute the C1-P2 for each receiver
      double d1 = rx1obs[C1].data - rx1obs[P2].data;
      double d2 = rx2obs[C1].data - rx2obs[P2].data;

      // Now compute the double differences
      res[prn] = d1 - d2;
   }

   return true;
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
PrnElevationMap elevation_map(const RODEpochMap& rem,
                              const gpstk::RinexObsHeader& roh,
                              const gpstk::EphemerisStore& eph)
{
   PrnElevationMap pem;
   if (verbosity>1)
      cout << "Computing elevation map." << endl;

   gpstk::ECEF rxpos(roh.antennaPosition);

   RODEpochMap::const_iterator i;
   for (i=rem.begin(); i!=rem.end(); i++)
   {
      const gpstk::DayTime& t = i->first;
      const gpstk::RinexObsData& rod = i->second;
      gpstk::RinexObsData::RinexPrnMap::const_iterator rpi;
      for (rpi=rod.obs.begin(); rpi!=rod.obs.end(); rpi++)
         try
         {
            short prn = rpi->first.prn;
            gpstk::Xvt svpos = eph.getPrnXvt(prn, t);
            pem[t][prn] = rxpos.elvAngle(svpos.x);
         }
         catch (gpstk::EphemerisStore::NoEphemerisFound& e)
         {
            if (verbosity>2)
               cout << e.getText() << endl;
         }
   }
   return pem;
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
PrnElevationMap elevation_map(const ORDEpochMap& oem)
{
   PrnElevationMap pem;
   if (verbosity>1)
      cout << "Computing elevation map." << endl;
   ORDEpochMap::const_iterator i;
   for (i=oem.begin(); i!=oem.end(); i++)
   {
      const gpstk::DayTime& t=i->first;
      const gpstk::ORDEpoch& epoch=i->second;
      gpstk::ORDEpoch::ORDMap::const_iterator j;
      for (j=epoch.ords.begin(); j!=epoch.ords.end(); j++)
      {
         int prn = j->first;
         gpstk::ObsRngDev ord = j->second;
         pem[t][prn] = ord.getElevation();
      }
   }
   return pem;
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void add_clock_to_rinex(RODEpochMap& rem, const ORDEpochMap& oem)
{
   const gpstk::GPSGeoid gm;
   RODEpochMap::iterator i;
   for (i=rem.begin(); i!=rem.end(); i++)
   {
      gpstk::RinexObsData& rod = i->second;

      ORDEpochMap::const_iterator j = oem.find(rod.time);
      if (j==oem.end() || !j->second.validClock)
      {
         if (verbosity)
            cout << "Epoch has no clock " << rod.time << endl;
         continue;
      }
      rod.clockOffset = j->second.clockOffset/gm.c();
   }
} // end of add_clock_to_rinex()




//-----------------------------------------------------------------------------
// This code both computes the bias on the phase, removes it, and detects
// discontinuities in the phase.
//-----------------------------------------------------------------------------
CycleSlipList debiasPhase(const RinexObsType rot, DDEpochMap& ddem)
{
   CycleSlipList sl;
   typedef std::pair<gpstk::RinexPrn, gpstk::RinexPrn> PrnPair;

   double f=0;
   if (rot==L1)
      f=gpstk::L1_FREQ; 
   else if (rot==L2)
      f=gpstk::L2_FREQ;
   else
      return sl;  //We only work for phase observables
   
   double lamda = gpstk::C_GPS_M/f;  // carrier wavelength
   double minSlipSize = 0.9 * lamda;

   if (verbosity>1)
      cout << "Minimum slip size = " << minSlipSize
           << " meters (wavelength = " << lamda
           << " meters)" << endl;

   // We scope this so the bm object gets flushed at the end of this scope
   {
      DDEpochMap::iterator ei;
      DDEpoch::DDResidualMap::iterator pi;
      map< PrnPair, gpstk::TwoSampleStats<double> > bm;
   
      for (ei = ddem.begin(); ei != ddem.end(); ei++)
         for (pi = ei->second.res.begin(); pi != ei->second.res.end(); pi++)
         {
            const gpstk::DayTime& t = ei->first;
            const gpstk::RinexPrn& prn = pi->first;
            DDResidual& ddr = pi->second;
            PrnPair pr(prn, ei->second.masterPrn);

            // does it looks like we need to compute a new bias?
            if (bm[pr].N() == 0 || abs(ddr.res[rot] - bm[pr].AverageY()) > minSlipSize)
               bm[pr].Reset();

            bm[pr].Add(t.MJDdate(), ddr.res[rot]);
            ddr.bias[rot]=bm[pr].AverageY();
         }
   }

   {
      DDEpochMap::reverse_iterator ei;
      DDEpoch::DDResidualMap::iterator pi;
      map< PrnPair, long double> bm;
      for (ei = ddem.rbegin(); ei != ddem.rend(); ei++)
         for (pi = ei->second.res.begin(); pi != ei->second.res.end(); pi++)
         {
            const gpstk::DayTime& t = ei->first;
            const gpstk::RinexPrn& prn = pi->first;
            DDResidual& ddr = pi->second;
            PrnPair pr(prn, ei->second.masterPrn);
            
            if (fabs(ddr.bias[rot] - bm[pr]) > minSlipSize)
            {
               if (bm[pr] != 0)
               {   
                  CycleSlipRecord s;
                  s.t = t;
                  s.cycles = (ddr.bias[rot]-bm[pr])/lamda;
                  s.prn = prn;
                  s.masterPrn =ei->second.masterPrn;
                  sl.push_front(s);
               }
               bm[pr] = ddr.bias[rot];
            }
            
            ddr.bias[rot] = bm[pr];
            ddr.res[rot] -= ddr.bias[rot];
         }
   }

   return sl;
}  // end of debiasPhase()



// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void dumpStats(const ORDEpochMap& oem, 
               const string& ordMode, 
               const double sigmam)
{
   if (verbosity) 
      cout << endl
           << "ord    elev    sdev     mean    z   #obs  #del   max   strip" << endl
           << "------ -----  -------  ----------  ------ ----  ------ ------" << endl;

   string desc = ordMode;

   computeStats(desc, oem, 0, 10, sigmam);
   computeStats(desc, oem, 10, 20, sigmam);
   computeStats(desc, oem, 20, 60, sigmam);
   computeStats(desc, oem, 60, 90, sigmam);
   computeStats(desc, oem, 10, 90, sigmam);

   if (verbosity>1)
      cout << endl
           << "sdev, mean, min, max, and strip in meters" << endl
           << "z: 0 if mean < sdev/sqrt(n)" << endl;
}

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void computeStats(
   const string desc,
   const ORDEpochMap& oem,
   const double minElevation,
   const double maxElevation,
   const double sigmam)
{
   gpstk::Stats<double> fp;
   ORDEpochMap::const_iterator ei;
   for (ei = oem.begin(); ei != oem.end(); ei++)
   {
      const gpstk::DayTime& t = ei->first;
      gpstk::ORDEpoch::ORDMap::const_iterator pi;
      for (pi = ei->second.ords.begin(); pi != ei->second.ords.end(); pi++)
      {
         const float el = pi->second.getElevation();
         const double ord = pi->second.getORD();
         if (el>minElevation && el<maxElevation && abs(ord) < 1e6)
            fp.Add(ord);
      }
   }

   double strip = sigmam * fp.StdDev();
   gpstk::Stats<double> good, bad;
   for (ei = oem.begin(); ei != oem.end(); ei++)
   {
      const gpstk::DayTime& t = ei->first;
      gpstk::ORDEpoch::ORDMap::const_iterator pi;
      for (pi = ei->second.ords.begin(); pi != ei->second.ords.end(); pi++)
      {
         const float el = pi->second.getElevation();
         const double ord = pi->second.getORD();
         if (el>minElevation && el<maxElevation)
         {
            if (abs(ord) < strip)
               good.Add(ord);
            else
               bad.Add(ord);
         }
      }
   }

   cout << left << setw(6) << desc << right
        << " "
        << setprecision(2) << setw(2) <<  minElevation
        << "-"
        << setw(2) << maxElevation
        << " ";

   char b1[200];
   char zero = good.Average() < good.StdDev()/sqrt((float)good.N())?'0':' ';
   double maxOrd = max(abs(good.Minimum()), abs(good.Maximum()));
   sprintf(b1, "%8.5f  %8.4f %c %7d %4d  %6.2f %6.2f",
           good.StdDev(), good.Average(), zero,
           good.N(), bad.N(), maxOrd, strip);
   
   cout << b1 << endl;
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void dump(std::ostream& s, const ORDEpochMap& oem)
{
   s.setf(ios::fixed, ios::floatfield);
   s << "# " << timeFormat
     << ", PRN, ORD(m), SV_EL, RX_CLK(m), iodc, health" << endl;

   ORDEpochMap::const_iterator ei;
   for (ei = oem.begin(); ei != oem.end(); ei++)
   {
      const gpstk::DayTime& t = ei->first;
      string time=t.printf(timeFormat);
      gpstk::ORDEpoch::ORDMap::const_iterator pi;
      const gpstk::ORDEpoch& e = ei->second;
      for (pi = e.ords.begin(); pi != e.ords.end(); pi++)
      {
         const short& prn = pi->first;
         const gpstk::ObsRngDev& ord = pi->second;
         
         s << dec << time
           << "   " << setw(2) << prn
           << setfill(' ')
           << "  " << setw(12) << setprecision(5) << ord.getORD()
           << "  " << setw(4)  << setprecision(1) << ord.getElevation()
           << "  " << setw(11) << setprecision(3) << e.clockOffset
           << hex
           << "  " << ord.getIODC()
           << "  " << ord.getHealth()
           << endl;
      }
   }
}  // end dump()



// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void dumpStats(DDEpochMap& ddem, PrnElevationMap& pem)
{
   if (verbosity) 
      cout << endl
           << "obs       elev    sdev(m)  mean(m) [err]         #obs   #del   min     max    " << endl
           << "--------- ------  -------  --------------------  ------ ----   -------------- " << endl;

   cout << "C1 dd res "; computeStats(C1, ddem,  0, 10, pem); cout << endl;
   cout << "C1 dd res "; computeStats(C1, ddem, 10, 20, pem); cout << endl;
   cout << "C1 dd res "; computeStats(C1, ddem, 20, 60, pem); cout << endl;
   cout << "C1 dd res "; computeStats(C1, ddem, 60, 90, pem); cout << endl;
   cout << "C1 dd res "; computeStats(C1, ddem, 10, 90, pem); cout << endl;
   cout << endl;
   cout << "P1 dd res "; computeStats(P1, ddem,  0, 10, pem); cout << endl;
   cout << "P1 dd res "; computeStats(P1, ddem, 10, 20, pem); cout << endl;
   cout << "P1 dd res "; computeStats(P1, ddem, 20, 60, pem); cout << endl;
   cout << "P1 dd res "; computeStats(P1, ddem, 60, 90, pem); cout << endl;
   cout << "P1 dd res "; computeStats(P1, ddem, 10, 90, pem); cout << endl;
   cout << endl;
   cout << "L1 dd res "; computeStats(L1, ddem,  0, 10, pem); cout << endl;
   cout << "L1 dd res "; computeStats(L1, ddem, 10, 20, pem); cout << endl;
   cout << "L1 dd res "; computeStats(L1, ddem, 20, 60, pem); cout << endl;
   cout << "L1 dd res "; computeStats(L1, ddem, 60, 90, pem); cout << endl;
   cout << "L1 dd res "; computeStats(L1, ddem, 10, 90, pem); cout << endl;
   cout << endl;
   cout << "P2 dd res "; computeStats(P2, ddem,  0, 10, pem); cout << endl;
   cout << "P2 dd res "; computeStats(P2, ddem, 10, 20, pem); cout << endl;
   cout << "P2 dd res "; computeStats(P2, ddem, 20, 60, pem); cout << endl;
   cout << "P2 dd res "; computeStats(P2, ddem, 60, 90, pem); cout << endl;
   cout << "P2 dd res "; computeStats(P2, ddem, 10, 90, pem); cout << endl;
   cout << endl;
   cout << "L2 dd res "; computeStats(L2, ddem,  0, 10, pem); cout << endl;
   cout << "L2 dd res "; computeStats(L2, ddem, 10, 20, pem); cout << endl;
   cout << "L2 dd res "; computeStats(L2, ddem, 20, 60, pem); cout << endl;
   cout << "L2 dd res "; computeStats(L2, ddem, 60, 90, pem); cout << endl;
   cout << "L2 dd res "; computeStats(L2, ddem, 10, 90, pem); cout << endl;

   cout << endl;
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void computeStats(
   const RinexObsType rot,
   DDEpochMap& ddem,
   const double minElevation,
   const double maxElevation,
   PrnElevationMap& pem)
{
   gpstk::Stats<double> good, bad;
   DDEpochMap::iterator ei;
   for (ei = ddem.begin(); ei != ddem.end(); ei++)
   {
      const gpstk::DayTime& t = ei->first;
      DDEpoch::DDResidualMap::iterator pi;
      for (pi = ei->second.res.begin(); pi != ei->second.res.end(); pi++)
      {
         const gpstk::RinexPrn& prn = pi->first;
         DDResidual& ddr = pi->second;
         if (pem[t][prn.prn]>minElevation && pem[t][prn.prn]<maxElevation)
         {
            if (abs(ddr.res[rot]) < 1000 )
               good.Add(ddr.res[rot]);
            else
               bad.Add(ddr.res[rot]);
         }
      }
   }
   
   cout << setprecision(2) << setw(2) <<  minElevation
        << ".."
        << setw(2) << maxElevation
        << " ";

   char b1[200];
   sprintf(b1, "%8.5f  %9.2e [%8.2e] %7d %4d  %7.2f %7.2f",
           good.StdDev()/2, good.Average(), good.StdDev()/sqrt((float)good.N()),
           good.N(), bad.N(),
           good.Minimum(), good.Maximum());

   cout << b1;
}

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void dumpStats(DD2EpochMap& ddem, PrnElevationMap& pem)
{
   if (verbosity)
      cout << endl
           << "obs       elev    sdev(m)  mean(m) [err]         #obs   #del   min     max    " << endl
           << "--------- ------  -------  --------------------  ------ ----   -------------- " << endl;

   cout << "C1-P2 res "; computeStats(ddem,  0, 10, pem); cout << endl;
   cout << "C1-P2 res "; computeStats(ddem, 10, 20, pem); cout << endl;
   cout << "C1-P2 res "; computeStats(ddem, 20, 60, pem); cout << endl;
   cout << "C1-P2 res "; computeStats(ddem, 60, 90, pem); cout << endl;
   cout << "C1-P2 res "; computeStats(ddem, 10, 90, pem); cout << endl;
   cout << endl;
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void computeStats(
   DD2EpochMap& ddem,
   const double minElevation,
   const double maxElevation,
   PrnElevationMap& pem)
{
   gpstk::Stats<double> good, bad;
   DD2EpochMap::iterator ei;
   for (ei = ddem.begin(); ei != ddem.end(); ei++)
   {
      const gpstk::DayTime& t = ei->first;
      DD2Epoch::DD2ResidualMap::iterator pi;
      for (pi = ei->second.res.begin(); pi != ei->second.res.end(); pi++)
      {
         const gpstk::RinexPrn& prn = pi->first;
         double ddr = pi->second;
         if (pem[t][prn.prn]>minElevation && pem[t][prn.prn]<maxElevation)
         {
            if (abs(ddr) < 1000 )
               good.Add(ddr);
            else
               bad.Add(ddr);
         }
      }
   }
   
   cout << setprecision(2) << setw(2) <<  minElevation
        << ".."
        << setw(2) << maxElevation
        << " ";

   char b1[200];
   sprintf(b1, "%8.5f  %9.2e [%8.2e] %7d %4d  %7.2f %7.2f",
           good.StdDev()/2, good.Average(), good.StdDev()/sqrt((float)good.N()),
           good.N(), bad.N(),
           good.Minimum(), good.Maximum());

   cout << b1;
}

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void dumpStats(const CycleSlipList& csl, const RinexObsType rot, PrnElevationMap& pem)
{
   if (csl.size()==0) return;
   cout << "Cycle slips detected on " << rot.type << " " << csl.size() << endl;

   if (verbosity<2) return;
   return;
   cout.setf(ios::fixed, ios::floatfield);
   CycleSlipList::const_iterator i;
   for (i=csl.begin(); i!=csl.end(); i++)
   {
      cout << "CS: " << i->t
           << "  "  << setw(8) << setprecision(1) << (i->t).GPSsecond()
           << "  " << setw(2) << i->masterPrn.prn << ":" << setw(2) << i->prn.prn
           << "  " << setprecision(4) << setw(12) << i->cycles
           << "  " << std::setprecision(3) << pem[i->t][i->prn.prn]
           << endl;
   }
}


// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void dump(std::ostream& s,
          DDEpochMap& ddem,
          const RinexObsType rot,
          PrnElevationMap& pem)
{
   s.setf(ios::fixed, ios::floatfield);
   s << "# " << timeFormat
     << ", PRN, RES(m), SV_EL, RX_CLK(ns)" << endl;

   DDEpochMap::iterator ei;
   DDEpoch::DDResidualMap::iterator pi;
   for (ei = ddem.begin(); ei != ddem.end(); ei++)
   {
      const gpstk::DayTime& t = ei->first;
      double clk = ei->second.clockOffset;
      string time=t.printf(timeFormat);
      for (pi = ei->second.res.begin(); pi != ei->second.res.end(); pi++)
      {
         const gpstk::RinexPrn& prn = pi->first;
         DDResidual& ddr = pi->second;
         s << time
           << setfill(' ')
           << " " << setw(2)  << prn.prn
           << setprecision(6)
           << " " << setw(11) << ddr.res[rot]
           << setprecision(1)
           << " " << setw(4) << pem[t][prn.prn]
           << " " << setw(11) << clk*1e9
           << endl;
      }
   }
}  // end dump()

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
void dump(std::ostream& s,
          DD2EpochMap& ddem,
          PrnElevationMap& pem)
{
   s.setf(ios::fixed, ios::floatfield);
   s << "# " << timeFormat 
     << ", PRN, RES(m), SV_EL" << endl;

   DD2EpochMap::iterator ei;
   DD2Epoch::DD2ResidualMap::iterator pi;
   for (ei = ddem.begin(); ei != ddem.end(); ei++)
   {
      const gpstk::DayTime& t = ei->first;
      string time=t.printf(timeFormat);
      for (pi = ei->second.res.begin(); pi != ei->second.res.end(); pi++)
      {
         const gpstk::RinexPrn& prn = pi->first;
         double ddr = pi->second;
         s << time
           << setfill(' ')
           << " " << setw(2)  << prn.prn
           << setprecision(6)
           << " " << setw(11) << ddr
           << setprecision(1)
           << " " << setw(4) << pem[t][prn.prn]
           << endl;
      }
   }
}  // end dump()
