#pragma ident "$Id: //depot/sgl/gpstk/dev/src/EngAlmanac.cpp#1 $"

//============================================================================
//
//  This file is part of GPSTk, the GPS Toolkit.
//
//  The GPSTk is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or
//  any later version.
//
//  The GPSTk is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with GPSTk; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  
//  Copyright 2004, The University of Texas at Austin
//
//============================================================================

//============================================================================
//
//This software developed by Applied Research Laboratories at the University of
//Texas at Austin, under contract to an agency or agencies within the U.S. 
//Department of Defense. The U.S. Government retains all rights to use,
//duplicate, distribute, disclose, or release this software. 
//
//Pursuant to DoD Directive 523024 
//
// DISTRIBUTION STATEMENT A: This software has been approved for public 
//                           release, distribution is unlimited.
//
//=============================================================================






/**
 * @file EngAlmanac.cpp
 * Almanac data encapsulated in engineering terms
 */

#include "gps_constants.hpp"
#include "DayTime.hpp"
#include "EngAlmanac.hpp"

using namespace std;
using namespace gpstk;

// This is a macro in order to retain useful location information in the exc
#define CHECK_SV_HERE(itty, prn) \
if (itty == almPRN.end()) \
{ \
   SVNotPresentException \
      exc("Attempt to get data from EngAlmData from a SV that is not" \
          " present."); \
   GPSTK_THROW(exc); \
}

namespace gpstk
{
   EngAlmanac :: EngAlmanac()
      throw()
   {
      for (int n = 0; n < 4; n++)
      {
         alpha[n] = beta[n] = 0.0;
      }
      
      A0 = A1 = dt_ls = dt_lsf = 0.0;
      
      t_ot = t_oa = 0;
      
      wn_t = wn_lsf = 0;
      
      alm_wk = 0;
      
      dn = 0;

      haveUTC = false;
   }

   bool EngAlmanac::addSubframe(const long subframe[10], const long gpsWeek)
      throw(InvalidParameter)
   {
      double ficked[60];
      short i, prn;

      if (!subframeConvert(subframe, gpsWeek, ficked))
         return false;

      short pat = getSubframePattern(subframe);

         // check tlm preamble, subframe id and format #
      if ((ficked[0] != 0x8b) || ((ficked[4] != 4) && (ficked [4] != 5))
          || (pat < 4) || (pat > 10))
      {
         InvalidParameter
            exc("EngAlmanac::addSubframe: Not a valid almanac page.");
         GPSTK_THROW(exc);
      }

      switch(pat)
      {
         case 4:
               /* Page with Orbital Elements */
               /*check PRN */
            if ((ficked[19] < 1) || (ficked[19] > MAX_PRN))
            {
               InvalidParameter exc("EngAlmData::addSF, PRN out of range");
               GPSTK_THROW(exc);
            }
            prn = (short)ficked[19];
            almPRN[prn] = AlmOrbit(prn, ficked[7], ficked[9], ficked[10],
                                   ficked[12], ficked[13], ficked[14],
                                   ficked[15], ficked[16], ficked[17],
                                   (long)ficked[8], (long)ficked[2], gpsWeek,
                                   (short)ficked[11]);
            break;

         case 5:  /*  Page with Satellite health information 1-24 */
            for(i=0; i < 24; i++)
               health[i] = (char)ficked[8 + i];
               // set the t_oa for this record
            t_oa = ((subframe[2] >> 14) & 0xFF) * 4096;
            break;

         case 6: 
         case 7:
               /* ignore page becase no data for our structure */
            return true;

         case 8: /* Page with UTC and ionosphere parameters */
            alpha[0] = ficked[7];
            alpha[1] = ficked[8];
            alpha[2] = ficked[9];
            alpha[3] = ficked[10];
            beta[0] = ficked[11];
            beta[1] = ficked[12];
            beta[2] = ficked[13];
            beta[3] = ficked[14];
            A0 = ficked[15];
            A1 = ficked[16];
            dt_ls = ficked[19];
            t_ot = (long)ficked[17];
            wn_t = (int)ficked[18];
            wn_lsf = (int)ficked[20];
            dn = (char)ficked[21];
            dt_lsf = ficked[22];
            haveUTC = true;
            break;

         case 9:  /* Page with Health for 25-32 and AS/SV config */
            for(i=0; i<MAX_PRN; i++)
               SV_config[i] = (char)ficked[7 + i];

            for(i=24; i<MAX_PRN; i++)
               health[i] = (char)ficked[15 + i];
            break;

         case 10:/* Page with Special Message */
            special_msg = "";
            for(i=0; i<22; i++)
               special_msg += (char)ficked[7 + i];
            break;

         default: 
               // never reached, see if statement prior to this switch
            break;
      }

      return true;
   }
   
   double EngAlmanac::getEcc(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of the ecc for the given PRN
      return (*i).second.ecc;
   }
   
   double EngAlmanac::getIOffset(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of the iOffset for the given PRN
      return (*i).second.i_offset;
   }

   double EngAlmanac::getOmegadot(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of OMEGAdot for the given PRN
      return (*i).second.OMEGAdot;
   }

   short EngAlmanac::getSVHealth(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of SV_health for the given PRN
      return (*i).second.SV_health;
   }

   double EngAlmanac::getAhalf(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of Ahalf for the given PRN
      return (*i).second.Ahalf;
   }

   double EngAlmanac::getA(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of A for the given PRN
      return (*i).second.Ahalf * (*i).second.Ahalf;
   }

   double EngAlmanac::getOmega0(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of OMEGA0 for the given PRN
      return (*i).second.OMEGA0;
   }

   double EngAlmanac::getW(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of w for the given PRN
      return (*i).second.w;
   }

   double EngAlmanac::getM0(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of M0 for the given PRN
      return (*i).second.M0;
   }

   double EngAlmanac::getAf0(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of the af0 for the given PRN
      return (*i).second.AF0;
   }
   
   
   double EngAlmanac::getAf1(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of af1 for the given PRN
      return (*i).second.AF1;
   }
   
   
   double EngAlmanac::getToa() const throw()
   {
      return (double)t_oa;
   }
   
   double EngAlmanac::getToa(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of the Toa for the given PRN
      return (double) (*i).second.Toa;
   }
   
   
   double EngAlmanac::getXmitTime(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of the xmit_time for the given PRN
      return (double) (*i).second.xmit_time;
   }
   
   
   short EngAlmanac::getFullWeek(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
      return (*i).second.getFullWeek();
   }

   void EngAlmanac::getIon(double a[4], double b[4]) const
      throw(InvalidRequest)
   {
      if (!haveUTC)
      {
         InvalidRequest 
            exc("UTC offset (subframe 4, page 18) is not present.");
         GPSTK_THROW(exc);
      }
      for (int n = 0; n < 4; n++)
      {
         a[n] = alpha[n];
         b[n] = beta[n];
      }
   }
      
   void EngAlmanac::getUTC(double& a0, double& a1, double& deltaTLS,
                           long& tot, int& WNt, int& WNLSF,
                           int& DN, double& deltaTLSF) const
      throw(InvalidRequest)
   {
      if (!haveUTC)
      {
         InvalidRequest 
            exc("UTC offset (subframe 4, page 18) is not present.");
         GPSTK_THROW(exc);
      }
      a0 = A0;
      a1 = A1;
      deltaTLS = dt_ls;
      tot = t_ot;
      WNt = wn_t;
      WNLSF = wn_lsf;
      DN = (int)dn;
      deltaTLSF = dt_lsf;
   }

   short EngAlmanac::getAlmWeek() const throw()
   {
      return alm_wk;
   }
   
   AlmOrbit EngAlmanac::getAlmOrbElem(short prn) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of the orbit elm. for the given PRN
      return (*i).second;
   }

   Xvt EngAlmanac::svXvt(short prn, const DayTime& t) const
      throw(EngAlmanac::SVNotPresentException)
   {
      AlmOrbits::const_iterator i = almPRN.find(prn);
      CHECK_SV_HERE(i, prn);
      
         // return value of the orbit elm. for the given PRN
      return (*i).second.svXvt(t);
   }

   bool EngAlmanac::isData(short prn) const throw()
   {
      return (almPRN.find(prn) != almPRN.end());
   }

} // namespace
