#pragma ident "$Id: //depot/sgl/gpstk/dev/src/Stats.hpp#2 $"

/**
 * @file Stats.hpp
 * One and two-sample statistics
 */
 
#ifndef GPSTK_STATS_HPP
#define GPSTK_STATS_HPP

//============================================================================
//
//  This file is part of GPSTk, the GPS Toolkit.
//
//  The GPSTk is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or
//  any later version.
//
//  The GPSTk is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with GPSTk; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  
//  Copyright 2004, The University of Texas at Austin
//
//============================================================================

#include "MiscMath.hpp"
#include "Vector.hpp"
#include "Exception.hpp"

namespace gpstk
{
   /** @addtogroup math */
   //@{

/** Conventional statistics for one sample.  Constructor does the same as
 * Reset(); use this when starting a new series of input samples.
 * Results are available at any time by calling N(), Minimum(), Maximum(),
 * Average(), Variance() and StdDev().
 */
   template <class T>
   class Stats
   {
   public:
      /// constructor
      explicit Stats() { n=0; }

      /// constructor given a vector of data
      Stats(Vector<T>& X, Vector<T>& W)
      {
         n = 0;
         Add(X,W);
      }

      /// reset, i.e. ignore earlier data and restart sampling
      inline void Reset(void) { n=0; }

      /// access the sample size
      inline unsigned int N(void) const { return n; }

      /// return minimum value
      inline T Minimum(void) const { if(n) return min; else return T(0); }

      /// return maximum value
      inline T Maximum(void) const { if(n) return max; else return T(0); }

      /// return computed average
      inline T Average(void) const
      {
         // normalization constant is (W=sum wts)/n, -> 1 when all wts=1
         if(n == 0) return T(0);
         if(W > T(0)) return (T(n)*ave/W);
         else return ave;
      }

      /// return computed variance
      inline T Variance(void) const
      {
         if(n == 0) return T(0);
         if(W > T(0)) {
            T wn=W/T(n);
            return (var/wn/wn/wn/wn);
         }
         else return var;
      }

      /// return computed standard deviation
      inline T StdDev(void) const
      {
         if(n == 0) return T(0);
         if(W > T(0)) {
            T wn=W/T(n);
            return SQRT(ABS(var))/wn/wn;
         }
         else return SQRT(ABS(var));
      }

      /// add a single sample to the computation of statistics, with optional weight
      void Add(const T x, const T wt=T(0))
      {
         T xx(x);
         n++;
         if(n == 1) {
            min = max = ave= xx;
            if(wt > T(0)) ave=wt*xx;
            var = T(0);
            W = T(0);
         }
         else {
            if(xx < min) min=xx;
            if(xx > max) max=xx;
         }

         if(wt > T(0)) { xx *= wt;  W += wt; }
         ave += (xx-ave)/T(n);
         if(n > 1) var = (var*T(n-2) + T(n)*(xx-ave)*(xx-ave)/T(n-1))/T(n-1);
      }

      /// add a Vector<T> of samples to the computation of statistics, with optional weights
      inline void Add(Vector<T>& X, Vector<T>& w)
      {
         for(size_t i=0; i<X.size(); i++) {
            if(w.size() > 0) 
               Add(X(i),w(i));
            else
               Add(X(i));
         }
      }

      /// remove a sample from the computation of statistics (can't do min and max).
      void Subtract(T x)
      {
         T dn=T(n);
         if(n > 2) var=(var*(dn-T(1))-dn*(x-ave)*(x-ave)/(dn-T(1)))/(dn-T(2));
         else var=T(0);

         if(n > 1) ave=(ave*dn-x)/(dn-T(1));
         else if(n==1) ave=x;
         else ave=T(0);

         n--;
      }

      /// remove a Vector<T> of samples to the computation of statistics
      inline void Subtract(Vector<T>& X)
      {
         for(size_t i=0; i<X.size(); i++) Subtract(X(i));
      }

      /// combine two Stats (assumed taken from the same or equivalent samples); both must be either weighted or unweighted.
      Stats<T>& operator+=(Stats<T>& S)
      {
         if(n + S.n == 0) return *this;
         if( (W==T(0) && S.W!=T(0)) || (W!=T(0) && S.W==T(0)) ) {
            gpstk::Exception e("Stats::operator+= : operands have inconsistent weighting");
            GPSTK_THROW(e);
         }

         if(S.min < min) min=S.min;
         if(S.max > max) max=S.max;

         T n1(n), n2(S.n);
         T oldave=ave;
         T W(1),W1(1),W2(1);

         if(W != T(0)) {               // weighted
            W1 = W/n1;
            W2 = S.W/n2;
            W = (W+S.W)/(n1+n2);
         }

         n += S.n;
         ave = (n1*W1*oldave + n2*W2*S.ave)/((n1+n2)*W);
         var  = W1*W1*((n1-T(1))*var + n1*oldave*oldave);
         var += W2*W2*((n2-T(1))*S.var + n2*S.ave*S.ave);
         var /= W*W;
         var -= (n1+n2)*ave*ave;
         var /= (n1+n2-T(1));

         return *this;

      }  // end Stats operator+=

   private:
      /// Number of samples added to the statistics so far
      unsigned int n;
      /// Minimum value
      T min;
      /// Maximum value
      T max;
      /// Average value
      T var;
      /// Variance (square of the standard deviation)
      T ave;
      /// Normalization constant = sum weights
      T W;
   }; // end class Stats

   /// Output operator for Stats class
   template <class T>
   std::ostream& operator<<(std::ostream& s, const Stats<T>& ST) 
   {
      std::ofstream savefmt;
      savefmt.copyfmt(s);
      s << " N       = " << ST.N() << "\n";
      s << " Minimum = "; s.copyfmt(savefmt); s << ST.Minimum();
      s << " Maximum = "; s.copyfmt(savefmt); s << ST.Maximum() << "\n";
      s << " Average = "; s.copyfmt(savefmt); s << ST.Average();
      s << " Std Dev = "; s.copyfmt(savefmt); s << ST.StdDev() << "\n";
      return s;
   }

/** Conventional statistics for two samples.  Constructor does the same as
 * Reset(); use this when starting a new series of input samples.
 * Results are available at any time by calling N(), Minimum(), Maximum(),
 * Average(), Variance() and StdDev().
 */
   template <class T>
   class TwoSampleStats
   {
   public:
      /// constructor
      TwoSampleStats() { n=0; }

      /// constructor given two Vector of data
      TwoSampleStats(Vector<T>& X, Vector<T>& Y)
      {
         n = 0;
         Add(X,Y);
      }

      /// Add data to the statistics
      void Add(const T& X, const T& Y)
      {
         if(n == 0) {
            sumx = sumy = sumx2 = sumy2 = sumxy = T(0);
            xmin = xmax = X;
            ymin = ymax = Y;
            scalex = scaley = T(1);
         }
         if(scalex==T(1) && X!=T(0)) scalex=ABS(X);
         if(scaley==T(1) && Y!=T(0)) scaley=ABS(Y);
         sumx += X/scalex;
         sumy += Y/scaley;
         sumx2 += (X/scalex)*(X/scalex);
         sumy2 += (Y/scaley)*(Y/scaley);
         sumxy += (X/scalex)*(Y/scaley);
         if(X < xmin) xmin=X;
         if(X > xmax) xmax=X;
         if(Y < ymin) ymin=Y;
         if(Y > ymax) ymax=Y;
         n++;
      }

      /// Add two Vectors of data to the statistics
      void Add(const Vector<T>& X, const Vector<T>& Y)
      {
         size_t m=(X.size()<Y.size()?X.size():Y.size());
         if(m==0) return;
         for(size_t i=0; i<m; i++) Add(X(i),Y(i));
      }

      void Subtract(const T& X, const T& Y)
      {
         if(n == 1) {
            sumx = sumy = sumx2 = sumy2 = sumxy = T(0);
            xmin = xmax = T(0);
            ymin = ymax = T(0);
            scalex = scaley = T(1);
            return;
         }

         sumx -= X/scalex;
         sumy -= Y/scaley;
         sumx2 -= (X/scalex)*(X/scalex);
         sumy2 -= (Y/scaley)*(Y/scaley);
         sumxy -= (X/scalex)*(Y/scaley);
         n--;
      }

      void Subtract(const Vector<T>& X, const Vector<T>& Y)
      {
         size_t m=(X.size()<Y.size()?X.size():Y.size());
         if(m==0) return;
         for(size_t i=0; i<m; i++) Subtract(X(i),Y(i));
      }

      /// reset, i.e. ignore earlier data and restart sampling
      inline void Reset(void) { n=0; }

      /// return the sample size
      inline unsigned int N(void) const { return n; }
      /// return minimum X value
      inline T MinimumX(void) const { if(n) return xmin; else return T(0); }
      /// return maximum X value
      inline T MaximumX(void) const { if(n) return xmax; else return T(0); }
      /// return minimum Y value
      inline T MinimumY(void) const { if(n) return ymin; else return T(0); }
      /// return maximum Y value
      inline T MaximumY(void) const { if(n) return ymax; else return T(0); }

      /// return computed X average
      inline T AverageX(void) const
         { if(n>0) return (scalex*sumx/T(n)); else return T(0); }

      /// return computed Y average
      inline T AverageY(void) const
         { if(n>0) return (scaley*sumy/T(n)); else return T(0); }

      /// return computed X variance
      inline T VarianceX(void) const
      {
         if(n>1) return scalex*scalex*(sumx2-sumx*sumx/T(n))/T(n-1);
         else return T(0);
      }

      /// return computed Y variance
      inline T VarianceY(void) const
      {
         if(n>1) return scaley*scaley*(sumy2-sumy*sumy/T(n))/T(n-1);
         else return T(0);
      }

      /// return computed X standard deviation
      inline T StdDevX(void) const { return SQRT(VarianceX()); }

      /// return computed Y standard deviation
      inline T StdDevY(void) const { return SQRT(VarianceY()); }

      /// return slope of best-fit line Y=slope*X + intercept
      inline T Slope(void) const
      {
         if(n>0) return ((scaley/scalex)*(sumxy-sumx*sumy/T(n))/(sumx2-sumx*sumx/T(n)));
         else return T(0);
      }

      /// return intercept of best-fit line Y=slope*X + intercept
      inline T Intercept(void) const
      {
         if(n>0) return (AverageY()-Slope()*AverageX());
         else return T(0);
      }

      /// return uncertainty in slope
      inline T SigmaSlope(void) const
      {
         if(n>2) return (SigmaYX()/(StdDevX()*SQRT(T(n-1))));
         else return T(0);
      }

      /// return correlation
      inline T Correlation(void) const
      {
         if(n>1) return ( scalex * scaley * (sumxy-sumx*sumy/T(n))
               / (StdDevX()*StdDevY()*T(n-1)) );
         else return T(0);
      }

      /// return conditional uncertainty = uncertainty y given x
      inline T SigmaYX(void) const
      {
         if(n>2) return (
            StdDevY() * SQRT(T(n-1)/T(n-2)) * SQRT(T(1)-Correlation()*Correlation()) );
         else return T(0);
      }

      /// combine two TwoSampleStats (assumed taken from the same or equivalent samples)
      TwoSampleStats<T>& operator+=(TwoSampleStats<T>& S)
      {
         if(n + S.n == 0) return *this;
         if(S.xmin < xmin) xmin=S.xmin;
         if(S.xmax > xmax) xmax=S.xmax;
         if(S.ymin < ymin) ymin=S.ymin;
         if(S.ymax > ymax) ymax=S.ymax;
         sumx += S.scalex*S.sumx/scalex;
         sumy += S.scaley*S.sumy/scaley;
         sumx2 += (S.scalex/scalex)*(S.scalex/scalex)*S.sumx2;
         sumy2 += (S.scaley/scaley)*(S.scaley/scaley)*S.sumy2;
         sumxy += (S.scalex/scalex)*(S.scaley/scaley)*S.sumxy;
         n += S.n;
         return *this;
      }  // end Stats operator+=

   private:
      /// Number of samples added to the statistics so far
      unsigned int n;
      T xmin, xmax, ymin, ymax, scalex, scaley;
      T sumx, sumy, sumx2, sumy2, sumxy;

   }; // end class TwoSampleStats

   /// Output operator for TwoSampleStats class
   template <class T>
   std::ostream& operator<<(std::ostream& s, const TwoSampleStats<T>& TSS) 
   {
      std::ofstream savefmt;
      savefmt.copyfmt(s);
      s << " N       = " << TSS.N() << "\n";
      s << " Minimum: X = "; s.copyfmt(savefmt); s << TSS.MinimumX();
      s << "  Y = "; s.copyfmt(savefmt); s << TSS.MinimumY();
      s << "  Maximum: X = "; s.copyfmt(savefmt); s << TSS.MaximumX();
      s << "  Y = "; s.copyfmt(savefmt); s << TSS.MaximumY() << "\n";
      s << " Average: X = "; s.copyfmt(savefmt); s << TSS.AverageX();
      s << "  Y = "; s.copyfmt(savefmt); s << TSS.AverageY();
      s << "  Std Dev: X = "; s.copyfmt(savefmt); s << TSS.StdDevX();
      s << "  Y = "; s.copyfmt(savefmt); s << TSS.StdDevY() << "\n";
      s << " Intercept = "; s.copyfmt(savefmt); s << TSS.Intercept();
      s << "  Slope = "; s.copyfmt(savefmt); s << TSS.Slope();
      s << " with uncertainty = "; s.copyfmt(savefmt); s << TSS.SigmaSlope() << "\n";
      s << " Conditional uncertainty (sigma y given x) = ";
      s.copyfmt(savefmt); s << TSS.SigmaYX();
      s << "  Correlation = "; s.copyfmt(savefmt); s << TSS.Correlation() << "\n";
      return s;
   }

   /** Compute the median of a vector */
   template <class T>
   inline T median(const Vector<T>& v)
   {
      if(v.size()==0) return T(0);
      if(v.size()==1) return v(0);
      if(v.size()==2) return (v(0)+v(1))/T(2);
      // insert sort
      size_t i,j;
      T x;
      Vector<T> w(v);
      for(i=0; i<v.size(); i++) {
         x = w[i] = v(i);
         j = i-1;
         while(j>=0 && x<w[j]) {
            w[j+1] = w[j];
            j--;
         }
         w[j+1] = x;
      }
      if(v.size() % 2)
         x=w[(v.size()+1)/2-1];
      else
         x=(w[v.size()/2-1]+w[v.size()/2])/T(2);
      delete[] w;
      return x;
   }  // end median(Vector)

   //@}

}  // namespace

#endif
